import numpy as np
import pytest

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.functions.function import FunctionError
from psyneulink.core.components.functions.learningfunctions import Hebbian, Reinforcement, TDLearning
from psyneulink.core.components.functions.objectivefunctions import Stability, Distance
from psyneulink.core.components.functions.distributionfunctions import NormalDist, ExponentialDist, \
    UniformDist, GammaDist, WaldDist, DriftDiffusionAnalytical
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import SimpleIntegrator, \
    AdaptiveIntegrator, DriftDiffusionIntegrator, OrnsteinUhlenbeckIntegrator, FitzHughNagumoIntegrator, \
    AccumulatorIntegrator, DualAdaptiveIntegrator
from psyneulink.core.components.functions.transferfunctions import Linear, Exponential, Logistic, SoftMax, LinearMatrix
from psyneulink.core.components.functions.combinationfunctions import Reduce, LinearCombination, CombineMeans
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism
from psyneulink.core.globals.keywords import \
    MAX_ABS_INDICATOR, MAX_ABS_ONE_HOT, MAX_ABS_VAL, MAX_INDICATOR, MAX_ONE_HOT, MAX_VAL, \
    MEAN, MEDIAN, PROB, STANDARD_DEVIATION, VARIANCE

class TestProcessingMechanismFunctions:

    @pytest.mark.benchmark(group="ProcessingMechanism[DefaultFunction]")
    @pytest.mark.parametrize("variable", [[1, 2, 3, 4],
                                          [1., 2., 3., 4.],
                                          np.asarray([1., 2., 3., 4.], dtype=np.int8),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int16),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int32),
                                          np.asarray([1., 2., 3., 4.], dtype=np.int64),
                                          np.asarray([1., 2., 3., 4.], dtype=np.float32),
                                          np.asarray([1., 2., 3., 4.], dtype=np.float64),
                                          [[1, 2, 3, 4]],
                                          [[1., 2., 3., 4.]],
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int8),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int16),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int32),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.int64),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.float32),
                                          np.asarray([[1., 2., 3., 4.]], dtype=np.float64),
                                         ],
                             ids=["list.int", "list.float", "np.1d.i8", "np.1d.i16", "np.1d.i32", "np.1d.i64", "np.1d.f32", "np.1d.f64",
                                  "list2d.int", "list2d.float", "np.2d.i8", "np.2d.i16", "np.2d.i32", "np.2d.i64", "np.2d.f32", "np.2d.f64",
                             ])
    @pytest.mark.parametrize("mode", ["Python",
                                      pytest.param("LLVM", marks=[pytest.mark.llvm]),
                                      pytest.param("PTX", marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                     ])
    def test_processing_mechanism_default_function(self, mode, variable, benchmark):
        PM = ProcessingMechanism(default_variable=[0, 0, 0, 0])
        if mode == "Python":
            ex = PM.execute
        elif mode == "LLVM":
            ex = pnlvm.MechExecution(PM).execute
        elif mode == "PTX":
            ex = pnlvm.MechExecution(PM).cuda_execute
        res = benchmark(ex, variable)
        assert np.allclose(res, [[1., 2., 3., 4.]])

    def test_processing_mechanism_linear_function(self):

        PM1 = ProcessingMechanism()
        PM1.execute(1.0)
        assert np.allclose(PM1.value, 1.0)

        PM2 = ProcessingMechanism(function=Linear(slope=2.0,
                                                  intercept=1.0))
        PM2.execute(1.0)
        assert np.allclose(PM2.value, 3.0)

    @pytest.mark.parametrize("function,expected", [(LinearCombination, [[1.]]),
                                                   (Reduce, [[1.]]),
                                                   (CombineMeans, [1.0]),
                                                   (Exponential, [[2.71828183]]),
                                                   (Logistic, [[0.73105858]]),
                                                   (SoftMax, [[1,]]),
                                                   (SimpleIntegrator, [[1.]]),
                                                   (AdaptiveIntegrator, [[1.]]),
                                                   (DriftDiffusionIntegrator, [[[1.]], [[1.]]]),
                                                   (OrnsteinUhlenbeckIntegrator, [[[-1.]], [[1.]]]),
                                                   (AccumulatorIntegrator, [[0.]]),
                                                   (FitzHughNagumoIntegrator, [[[0.05127053]], [[0.00279552]], [[0.05]]]),
                                                   (DualAdaptiveIntegrator, [[0.1517455]]),
                                                   (DriftDiffusionAnalytical, [[1.19932930e+00],
                                                                               [3.35350130e-04],
                                                                               [1.19932930e+00],
                                                                               [2.48491374e-01],
                                                                               [1.48291009e+00],
                                                                               [1.19932930e+00],
                                                                               [2.48491374e-01],
                                                                               [1.48291009e+00]]),
                                                   (NormalDist, [[-0.51529709]]),
                                                   (ExponentialDist, [[0.29964231]]),
                                                   (UniformDist, [[0.25891675]]),
                                                   (GammaDist, [[0.29964231]]),
                                                   (WaldDist, [[0.73955962]]),
                                                  ],
                             ids=lambda x: getattr(x, "componentName", ""))
    def test_processing_mechanism_function(self, function, expected):
        PM = ProcessingMechanism(function=function)
        res = PM.execute(1.0)
        assert np.allclose(np.asfarray(res), expected)

    # COMMENTED OUT BECAUSE OF MATLAB ENGINE:
    # def test_processing_mechanism_NavarroAndFuss_function(self):
    #     PM1 = ProcessingMechanism(function=NavarroAndFuss)
    #     PM1.execute(1.0)
    #     # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Distance_function(self):
        PM1 = ProcessingMechanism(function=Distance,
                                  default_variable=[[0,0], [0,0]])
        PM1.execute([[1, 2], [3, 4]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Hebbian_function(self):
        PM1 = ProcessingMechanism(function=Hebbian,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_Reinforcement_function(self):
        PM1 = ProcessingMechanism(function=Reinforcement,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    # COMMENTING OUT BECAUSE BACK PROP FN DOES NOT WORK WITH UNRESTRICTED MECHANISM
    # def test_processing_mechanism_BackPropagation_function(self):
    #     PM1 = ProcessingMechanism(function=BackPropagation,
    #                                          default_variable=[[0.0], [0.0], [0.0]])
    #     PM1.execute([[1.0], [2.0], [3.0]])
    #     PM1.execute(1.0)
    #     # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_TDLearning_function(self):
        PM1 = ProcessingMechanism(function=TDLearning,
                                  default_variable=[[0.0], [0.0], [0.0]])
        PM1.execute([[1.0], [2.0], [3.0]])
        # assert np.allclose(PM1.value, 1.0)

    def test_processing_mechanism_multiple_input_ports(self):
        PM1 = ProcessingMechanism(size=[4, 4], function=LinearCombination, input_ports=['input_1', 'input_2'])
        PM2 = ProcessingMechanism(size=[2, 2, 2], function=LinearCombination, input_ports=['1', '2', '3'])
        PM1.execute([[1, 2, 3, 4], [5, 4, 2, 2]])
        PM2.execute([[2, 0], [1, 3], [1, 0]])
        assert np.allclose(PM1.value, [6, 6, 5, 6])
        assert np.allclose(PM2.value, [4, 3])

class TestLinearMatrixFunction:

    def test_valid_matrix_specs(self):
        # Note: default matrix specification is None

        PM_default = ProcessingMechanism(function=LinearMatrix())
        PM_default.execute(1.0)

        assert np.allclose(PM_default.value, 1.0)

        PM_default_len_2_var = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0]])
        PM_default_len_2_var.execute([[1.0, 2.0]])

        assert np.allclose(PM_default_len_2_var.value, [[1.0, 2.0]])

        PM_default_2d_var = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0],
                                                                                        [0.0, 0.0],
                                                                                        [0.0, 0.0]]),
                                                   default_variable=[[0.0, 0.0],
                                                                     [0.0, 0.0],
                                                                     [0.0, 0.0]])

        #  [1.0   0.0]           [1.0   0.0]
        #  [0.0   2.0]    *      [0.0   1.0]
        #  [3.0   0.0]
        PM_default_2d_var.execute([[1.0, 0.0],
                                   [0.0, 2.0],
                                   [3.0, 0.0]])

        assert np.allclose(PM_default_2d_var.value, [[1.0, 0.0],
                                                     [0.0, 2.0],
                                                     [3.0, 0.0]])

        # PM_float = ProcessingMechanism(function=LinearMatrix(matrix=4.0))
        # PM_float.execute(1.0)
        #
        # assert np.allclose(PM_float.value, 4.0)

        PM_1d_list = ProcessingMechanism(function=LinearMatrix(matrix=[4.0]))
        PM_1d_list.execute(1.0)

        assert np.allclose(PM_1d_list.value, 4.0)

        PM_2d_list = ProcessingMechanism(function=LinearMatrix(matrix=[[4.0, 5.0],
                                                                       [6.0, 7.0],
                                                                       [8.0, 9.0],
                                                                       [10.0, 11.0]],
                                                               default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0],
                                                                                 [0.0, 0.0, 0.0, 0.0]]),
                                         default_variable=[[0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0],
                                                           [0.0, 0.0, 0.0, 0.0]])
        PM_2d_list.execute([[1.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 1.0, 0.0],
                            [0.0, 0.0, 0.0, 1.0]])

        assert np.allclose(PM_2d_list.value, [[4.0, 5.0],
                                              [8.0, 9.0],
                                              [10.0, 11.0]])

        PM_1d_array = ProcessingMechanism(function=LinearMatrix(matrix=np.array([4.0])))
        PM_1d_array.execute(1.0)

        assert np.allclose(PM_1d_array.value, 4.0)

        PM_2d_array = ProcessingMechanism(function=LinearMatrix(matrix=np.array([[4.0]])))
        PM_2d_array.execute(1.0)

        assert np.allclose(PM_2d_array.value, 4.0)

        PM_matrix = ProcessingMechanism(function=LinearMatrix(matrix=np.matrix([[4.0]])))
        PM_matrix.execute(1.0)

        assert np.allclose(PM_matrix.value, 4.0)

    def test_invalid_matrix_specs(self):

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_float = ProcessingMechanism(function=LinearMatrix(default_variable=0.0,
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=0.0)
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

        with pytest.raises(FunctionError) as error_text:
            PM_mismatched_matrix = ProcessingMechanism(function=LinearMatrix(default_variable=[[0.0, 0.0],
                                                                                          [0.0, 0.0],
                                                                                          [0.0, 0.0]],
                                                                        matrix=[[1.0, 0.0, 0.0, 0.0],
                                                                                [0.0, 2.0, 0.0, 0.0],
                                                                                [0.0, 0.0, 3.0, 0.0],
                                                                                [0.0, 0.0, 0.0, 4.0]]),
                                                  default_variable=[[0.0, 0.0],
                                                                    [0.0, 0.0],
                                                                    [0.0, 0.0]])
        assert "Specification of matrix and/or default_variable" in str(error_text.value) and \
               "not compatible for multiplication" in str(error_text.value)

class TestProcessingMechanismStandardOutputPorts:

    @pytest.mark.benchmark
    @pytest.mark.parametrize("op, expected", [(MAX_ONE_HOT, [0, 2, 0]),
                                              (MAX_INDICATOR, [0, 1, 0]),
                                              (MAX_ABS_INDICATOR, [0, 0, 1]),
                                             ],
                             ids=lambda x: x if isinstance(x, str) else "")
    @pytest.mark.parametrize("mode", ["Python",
                                      pytest.param("LLVM", marks=[pytest.mark.llvm]),
                                      pytest.param("PTX", marks=[pytest.mark.llvm, pytest.mark.cuda]),
                                     ])
    def test_output_ports(self, mode, op, expected, benchmark):
        benchmark.group = "Output Port Op: {}".format(op)
        PM1 = ProcessingMechanism(default_variable=[0, 0, 0], output_ports=[op])
        var = [1, 2, 4] if op in {MEAN, MEDIAN, STANDARD_DEVIATION, VARIANCE} else [1, 2, -4]
        if mode == "Python":
            ex = PM1.execute
        elif mode == "LLVM":
            ex = pnlvm.MechExecution(PM1).execute
        elif mode == "PTX":
            ex = pnlvm.MechExecution(PM1).cuda_execute
        res = benchmark(ex, var)
        res = PM1.output_ports[0].value if mode == "Python" else res
        assert np.allclose(res, expected)

    # FIXME: These variants don't compile (use UDFs)
    @pytest.mark.parametrize("op, expected", [(MEAN, [2.33333333]),
                                              (MEDIAN, [2]),
                                              (STANDARD_DEVIATION, [1.24721913]),
                                              (VARIANCE, [1.55555556]),
                                              (MAX_VAL, [2]),
                                              (MAX_ABS_VAL, [4]),
                                              (MAX_ABS_ONE_HOT, [0, 0, 4]),
                                              (PROB, [0, 2, 0]),
                                             ],
                             ids=lambda x: x if isinstance(x, str) else "")
    def test_output_ports2(self, op, expected):
        PM1 = ProcessingMechanism(default_variable=[0, 0, 0], output_ports=[op])
        var = [1, 2, 4] if op in {MEAN, MEDIAN, STANDARD_DEVIATION, VARIANCE} else [1, 2, -4]
        PM1.execute(var)
        assert np.allclose(PM1.output_ports[0].value, expected)



from collections.abc import Iterable
import numpy as np
import pytest

from psyneulink.core.globals.utilities import convert_all_elements_to_np_array, prune_unused_args


@pytest.mark.parametrize(
    'arr, expected',
    [
        ([[0], [0, 0]], np.array([np.array([0]), np.array([0, 0])], dtype=object)),
        # should test these but numpy cannot easily create an array from them
        # [np.ones((2,2)), np.zeros((2,1))]
        # [np.array([[0]]), np.array([[[ 1.,  1.,  1.], [ 1.,  1.,  1.]]])]

    ]
)
def test_convert_all_elements_to_np_array(arr, expected):
    converted = convert_all_elements_to_np_array(arr)

    # no current numpy methods can test this
    def check_equality_recursive(arr, expected):
        if (
            not isinstance(arr, Iterable)
            or (isinstance(arr, np.ndarray) and arr.ndim == 0)
        ):
            assert arr == expected
        else:
            assert isinstance(expected, type(arr))
            assert len(arr) == len(expected)

            for i in range(len(arr)):
                check_equality_recursive(arr[i], expected[i])

    check_equality_recursive(converted, expected)


def f():
    pass


def g(a):
    pass


def h(a, b=None):
    pass


def i(b=None):
    pass


@pytest.mark.parametrize(
    'func, args, kwargs, expected_pruned_args, expected_pruned_kwargs', [
        (f, 1, {'a': 1}, [], {}),
        (g, 1, {'x': 1}, [1], {}),
        (g, None, {'a': 1}, [], {'a': 1}),
        (h, None, {'a': 1, 'b': 1, 'c': 1}, [], {'a': 1, 'b': 1}),
        (h, [1, 2, 3], None, [1], {}),
        (i, None, {'a': 1, 'b': 1, 'c': 1}, [], {'b': 1}),
        (i, [1, 2, 3], None, [], {}),
    ]
)
def test_prune_unused_args(func, args, kwargs, expected_pruned_args, expected_pruned_kwargs):
    pruned_args, pruned_kwargs = prune_unused_args(func, args, kwargs)

    assert pruned_args == expected_pruned_args
    assert pruned_kwargs == expected_pruned_kwargs




# this scrip implements Gilbert and Shallice 2002 PDP Task Switching Model
import numpy as np
import psyneulink as pnl


### LAYERS
WORD_INPUT_LAYER = pnl.TransferMechanism(size = 3,
                                         function=pnl.Linear,
                                         name='WORD INPUT LAYER')

COLOR_INPUT_LAYER = pnl.TransferMechanism(size = 3,
                                         function=pnl.Linear,
                                         name='COLOR INPUT LAYER')

WORD_OUTPUT_LAYER = pnl.IntegratorMechanism(size = 3,
                                                   # auto= 0.0,
                                                   # hetero= -2.0,
                                                   function= pnl.InteractiveActivationIntegrator(decay= 0.0015, rest=-6),
                                                   name='WORD OUTPUT LAYER')
WORD_OUTPUT_LAYER.set_log_conditions('value')


COLOR_OUTPUT_LAYER = pnl.IntegratorMechanism(size = 3,
                                                   # auto= 0.0,
                                                   # hetero= -2.0,
                                                   function= pnl.InteractiveActivationIntegrator(decay= 0.0015, rest=-6, ),
                                             #  (rest= -6),
                                                   name='COLOR OUTPUT LAYER')

COLOR_OUTPUT_LAYER.set_log_conditions('value')


TASK_DEMAND_LAYER = pnl.IntegratorMechanism(size = 2,
                                                   # auto= 0.0,
                                                   # hetero= -2.0,
                                                   function= pnl.InteractiveActivationIntegrator(decay= 0.0015, max_val=1,
                                                                                       min_val= 1, rest= -4),
                                                   name='TASK DEMAND LAYER')

WORD_RECURRENT_LAYER = pnl.TransferMechanism(size = 3,
                                             function=pnl.Linear,
                                             name = 'WORD RECURRENT LAYER')

COLOR_RECURRENT_LAYER = pnl.TransferMechanism(size = 3,
                                             function=pnl.Linear,
                                             name = 'COLOR RECURRENT LAYER')

### WEIGHTS

#   WORD INPUT TO WORD OUTPUT
word_weights = pnl.MappingProjection(matrix=np.matrix([[3.5, 0.0, 0.0],
                                                        [0.0, 3.5, 0.0],
                                                        [0.0, 0.0, 3.5]]),
                                     name='WORD_WEIGHTS')

#   COLOR INPUT TO COLOR OUTPUT
color_weights = pnl.MappingProjection(matrix=np.matrix([[1.9, 0.0, 0.0],
                                                        [0.0, 1.9, 0.0],
                                                        [0.0, 0.0, 1.9]]),
                                      name='COLOR_WEIGHTS')

#    WORD INPUT to TASK DEMAND LAYER
word_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[1.0, 1.0],
                                                                    [1.0, 1.0],
                                                                    [1.0, 1.0]]),
                                                 name='WORD_TASK_DEMAND_WEIGHTS')

#    COLOR INPUT to TASK DEMAND LAYER
color_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[1.0, 1.0],
                                                                    [1.0, 1.0],
                                                                    [1.0, 1.0]]),
                                                  name='COLOR_TASK_DEMAND_WEIGHTS')


#   TASK DEMAND TO WORD OUTPUT
task_demand_word_output_weights = pnl.MappingProjection(matrix=np.matrix([[2.5, 2.5, 2.5],
                                                                          [-2.5, -2.5, -2.5]]),
                                                        name='TASK_DEMAND_WORD_OUTPUT_WEIGHTS')

#   TASK DEMAND TO COLOR OUTPUT
task_demand_color_output_weights = pnl.MappingProjection(matrix=np.matrix([[-2.5, -2.5, -2.5],
                                                                           [2.5, 2.5, 2.5]]),
                                                         name='TASK_DEMAND_COLOR_OUTPUT_WEIGHTS')

#   WORD OUTPUT TO TASK DEMAND
word_output_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[1.0, -1.0],
                                                                          [1.0, -1.0],
                                                                          [1.0, -1.0]]),
                                                        name='WORD_OUTPUT_TASK_DEMAND_WEIGHTS')

#   WORD OUTPUT TO TASK DEMAND
color_output_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[-1.0, 1.0],
                                                                           [-1.0, 1.0],
                                                                           [-1.0, 1.0]]),
                                                        name='COLOR_OUTPUT_TASK_DEMAND_WEIGHTS')

#   WORD OUTPUT to COLOR OUTPUT
word_output_color_output_weights = pnl.MappingProjection(matrix=np.matrix([[2.0, -2.0, -2.0],
                                                                           [-2.0, 2.0, -2.0],
                                                                           [-2.0, -2.0, 2.0]]),
                                                         name='WORD_OUTPUT_COLOR_OUTPUT_WEIGHTS')
# WORD OUTPUT TO TASK DEMAND
word_output_output_to_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[1.0, 1.0],
                                                                                          [1.0, 1.0],
                                                                                          [1.0, 1.0]]),
                                                                        name='WORD_COLOR_OUTPUT_TASK_DEMAND_WEIGHTS')

# COLOR OUTPUT TO TASK DEMAND
color_output_output_to_task_demand_weights = pnl.MappingProjection(matrix=np.matrix([[1.0, 1.0],
                                                                                          [1.0, 1.0],
                                                                                          [1.0, 1.0]]),
                                                                        name='COLOR_COLOR_OUTPUT_TASK_DEMAND_WEIGHTS')

# RECURRENT WORD weights
word_recurrent = pnl.MappingProjection(matrix=np.matrix([[0.0, -2.0, -2.0],
                                                         [-2.0, 0.0, -2.0],
                                                         [-2.0, -2.0, 0.0]]),
                                       name='WORD_RECURRENT_WEIGHTS')

# RECURRENT COLOR weights
color_recurrent = pnl.MappingProjection(matrix=np.matrix([[0.0, -2.0, -2.0],
                                                          [-2.0, 0.0, -2.0],
                                                          [-2.0, -2.0, 0.0]]),
                                       name='TASK_RECURRENT_WEIGHTS')

# RECURRENT TASK weights
task_recurrent = pnl.MappingProjection(matrix=np.matrix([[0.0, -2.0, -2.0],
                                                         [-2.0, 0.0, -2.0],
                                                         [-2.0, -2.0, 0.0]]),
                                       name='TASK_RECURRENT_WEIGHTS')

### COMPOSITION

Gilbert_Shallice_System = pnl.Composition(name="Gilbert_Shallice_System")

### ADD pathways
### pathway word input word output
words_input_output_pathway = [WORD_INPUT_LAYER,
                              word_weights,
                              WORD_OUTPUT_LAYER]

Gilbert_Shallice_System.add_linear_processing_pathway(pathway=words_input_output_pathway)

### pathway color input color output
color_input_output_pathway = [COLOR_INPUT_LAYER,
                              color_weights,
                              COLOR_OUTPUT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=color_input_output_pathway)

# PATHWAY: TASK WORD OUTPUT
task_word_output_pathway = [TASK_DEMAND_LAYER,
                              task_demand_word_output_weights,
                              WORD_OUTPUT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=task_word_output_pathway)

# Pathway: Task demand color output pathway
task_color_output_pathway = [TASK_DEMAND_LAYER,
                             task_demand_color_output_weights,
                             COLOR_OUTPUT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=task_color_output_pathway)

### Pathway: word input task demand
word_input_task_pathway = [WORD_INPUT_LAYER,
                           word_task_demand_weights,
                           TASK_DEMAND_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=word_input_task_pathway)

### Pathway: color input task demand
color_input_task_pathway = [COLOR_INPUT_LAYER,
                            color_task_demand_weights,
                            TASK_DEMAND_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=color_input_task_pathway)

### Pathway: Word output Task
word_output_task_pathway = [WORD_OUTPUT_LAYER,
                            word_output_output_to_task_demand_weights,
                            TASK_DEMAND_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=word_output_task_pathway)

### Pathway: Color output Task
color_output_task_pathway = [COLOR_OUTPUT_LAYER,
                             color_output_output_to_task_demand_weights,
                             TASK_DEMAND_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=color_output_task_pathway)

### Pathway: word output - recurrent word
word_output_word_recurrent_pathway = [WORD_OUTPUT_LAYER,
                                       WORD_RECURRENT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=word_output_word_recurrent_pathway)

### Pathway: color output - recurrent color
color_output_color_recurrent_pathway = [COLOR_OUTPUT_LAYER,
                                       COLOR_RECURRENT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=color_output_color_recurrent_pathway)


### Pathway: recurrent word - word output
word_recurrent_word_output_pathway = [WORD_RECURRENT_LAYER,
                                      word_recurrent,
                                      WORD_OUTPUT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=word_recurrent_word_output_pathway,
                                                      feedback=True)

### Pathway:  recurrent color - color output
color_recurrent_color_output_pathway = [COLOR_RECURRENT_LAYER,
                                        color_recurrent,
                                        COLOR_OUTPUT_LAYER]
Gilbert_Shallice_System.add_linear_processing_pathway(pathway=color_recurrent_color_output_pathway,
                                                      feedback=True)


## specify terminal layers (can only specify one terminal layer at a time)
# # MODIFIED 4/25/20 OLD:
# Gilbert_Shallice_System.add_required_node_role(WORD_OUTPUT_LAYER, pnl.NodeRole.TERMINAL)
# Gilbert_Shallice_System.add_required_node_role(COLOR_OUTPUT_LAYER, pnl.NodeRole.TERMINAL)
# MODIFIED 4/25/20 NEW:
Gilbert_Shallice_System.require_node_roles(WORD_OUTPUT_LAYER, pnl.NodeRole.TERMINAL)
Gilbert_Shallice_System.require_node_roles(COLOR_OUTPUT_LAYER, pnl.NodeRole.TERMINAL)
# MODIFIED 4/25/20 END

Gilbert_Shallice_System._analyze_graph()
# Gilbert_Shallice_System.show_graph()

input_dict = {COLOR_INPUT_LAYER: [1, 0, 0],
              WORD_INPUT_LAYER: [1, 0, 0],
              # TASK_DEMAND_LAYER: [1, 0]
              }

Gilbert_Shallice_System.run(num_trials=3,
                            inputs=input_dict)

WORD_OUTPUT_LAYER.log.print_entries()
COLOR_OUTPUT_LAYER.log.print_entries()

### SYSTEM



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# **************************************  IntegratorMechanism *************************************************

"""

Contents
--------

  * `IntegratorMechanism_Overview`
  * `IntegratorMechanism_Creation`
  * `IntegratorMechanism_Structure`
  * `IntegratorMechanism_Execution`
  * `IntegratorMechanism_Class_Reference`


.. _IntegratorMechanism_Overview:

Overview
--------

An IntegratorMechanism integrates its input, possibly based on its prior values.  The input can be a single
scalar value or an array of scalars (list or 1d np.array).  If it is a list or array, then each value is
independently integrated.  The default function (`IntegratorFunction`) can be parametrized to implement either a simple
increment rate, additive accumulator, or an (exponentially weighted) time-averaging of its input.  It can also be
assigned a custom function.

.. _IntegratorMechanism_Creation:

Creating an IntegratorMechanism
-------------------------------

An IntegratorMechanism can be created directly by calling its constructor, or using the `mechanism` command and
specifying *INTEGRATOR_MECHANISM* as its **mech_spec** argument.  Its function is specified in the **function**
argument, which can be parametrized by calling its constructor with parameter values::

    >>> import psyneulink as pnl
    >>> my_time_averaging_mechanism = pnl.IntegratorMechanism(function=pnl.AdaptiveIntegrator(rate=0.5))

The **default_variable** argument specifies the format of its input (i.e., whether it is a single scalar or an
array), as well as the value to use if none is provided when Mechanism is executed.  Alternatively, the **size**
argument can be used to specify the length of the array, in which case it will be initialized with all zeros.

.. _IntegratorMechanism_Structure:

Structure
---------

An IntegratorMechanism has a single `InputPort`, the `value <InputPort.InputPort.value>` of which is
used as the  `variable <Mechanism_Base.variable>` for its `function <IntegratorMechanism.function>`.
The default for `function <IntegratorMechanism.function>` is `AdaptiveIntegrator(rate=0.5)`. However,
a custom function can also be specified,  so long as it takes a numeric value, or a list or np.ndarray of numeric
values as its input, and returns a value of the same type and format.  The Mechanism has a single `OutputPort`,
the `value <OutputPort.OutputPort.value>` of which is assigned the result of  the call to the Mechanism's
`function  <IntegratorMechanism.function>`.

.. _IntegratorMechanism_Execution:

Execution
---------

When an IntegratorMechanism is executed, it carries out the specified integration, and assigns the
result to the `value <Mechanism_Base.value>` of its `primary OutputPort <OutputPort_Primary>`.  For the default
function (`IntegratorFunction`), if the value specified for **default_variable** is a list or array, or **size** is greater
than 1, each element of the array is independently integrated.  If its `rate <IntegratorFunction.rate>` parameter is a
single value,  that rate will be used for integrating each element.  If the `rate <IntegratorFunction.rate>` parameter is a
list or array, then each element will be used as the rate for the corresponding element of the input (in this case,
`rate <IntegratorFunction.rate>` must be the same length as the value specified for **default_variable** or **size**).


.. _IntegratorMechanism_Class_Reference:

Class Reference
---------------

"""
from collections.abc import Iterable

import typecheck as tc
import numpy as np

from psyneulink.core.components.functions.function import Function
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import AdaptiveIntegrator
from psyneulink.core.components.mechanisms.processing.processingmechanism import ProcessingMechanism_Base
from psyneulink.core.components.mechanisms.mechanism import Mechanism
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.keywords import \
    DEFAULT_VARIABLE, INTEGRATOR_MECHANISM, RESULTS, VARIABLE, PREFERENCE_SET_NAME
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set, REPORT_OUTPUT_PREF
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = [
    'DEFAULT_RATE', 'IntegratorMechanism', 'IntegratorMechanismError'
]

# IntegratorMechanism parameter keywords:
DEFAULT_RATE = 0.5

class IntegratorMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class IntegratorMechanism(ProcessingMechanism_Base):
    """
    IntegratorMechanism( \
        function=AdaptiveIntegrator(rate=0.5))

    Subclass of `ProcessingMechanism <ProcessingMechanism>` that integrates its input.
    See `Mechanism <Mechanism_Class_Reference>` for additional arguments and attributes.


    Arguments
    ---------

    function : IntegratorFunction : default IntegratorFunction
        specifies the function used to integrate the input.  Must take a single numeric value, or a list or np.array
        of values, and return one of the same form.

    """

    componentType = INTEGRATOR_MECHANISM

    classPreferenceLevel = PreferenceLevel.TYPE
    # These will override those specified in TYPE_DEFAULT_PREFERENCES
    classPreferences = {
        PREFERENCE_SET_NAME: 'IntegratorMechanismCustomClassPreferences',
        REPORT_OUTPUT_PREF: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ProcessingMechanism_Base.Parameters):
        """
            Attributes
            ----------

                function
                    see `function <IntegratorMechanism.function>`

                    :default value: `AdaptiveIntegrator`(initializer=numpy.array([0]), rate=0.5)
                    :type: `Function`
        """
        function = Parameter(AdaptiveIntegrator(rate=0.5), stateful=False, loggable=False)

        #
    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 size=None,
                 input_ports:tc.optional(tc.any(list, dict))=None,
                 function=None,
                 output_ports:tc.optional(tc.any(str, Iterable))=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):
        """Assign type-level preferences, default input value (SigmoidLayer_DEFAULT_BIAS) and call super.__init__
        """

        super(IntegratorMechanism, self).__init__(default_variable=default_variable,
                                                  size=size,
                                                  function=function,
                                                  params=params,
                                                  name=name,
                                                  prefs=prefs,
                                                  input_ports=input_ports,
                                                  output_ports=output_ports,
                                                  **kwargs)

        # IMPLEMENT: INITIALIZE LOG ENTRIES, NOW THAT ALL PARTS OF THE MECHANISM HAVE BEEN INSTANTIATED

    # def _parse_function_variable(self, variable, context=None, context=None):
    #     super()._parse_function_variable(variable, context, context)

    def _handle_default_variable(self, default_variable=None, size=None, input_ports=None, function=None, params=None):
        """If any parameters with len>1 have been specified for the Mechanism's function, and Mechanism's
        default_variable has not been specified, reshape Mechanism's variable to match function's,
        but make sure function's has the same outer dimensionality as the Mechanism's
        """

        # Get variable for Mechanism
        user_specified = False
        if default_variable is not None:
            variable = np.atleast_1d(default_variable)
            user_specified = True
        else:
            variable = self.parameters.variable.default_value
            user_specified = self.parameters.variable._user_specified

        # Only bother if an instantiated function was specified for the Mechanism
        if isinstance(function, Function):
            function_variable = function.parameters.variable.default_value
            function_variable_len = function_variable.shape[-1]
            variable_len = variable.shape[-1]

            # Raise error if:
            # - the length of both Mechanism and function variable are greater than 1 and they don't match, or
            # - the Mechanism's variable length is 1 and the function's is > 1 (in which case would like to assign
            #   shape of function's variable to that of Mechanism) but Mechanism's variable is user-specified.
            if ((variable_len>1 and function_variable_len>1 and variable_len!=function_variable_len) or
                (function_variable_len>1 and variable_len==1 and user_specified)):
                raise IntegratorMechanismError(f"Shape of {repr(VARIABLE)} for function specified for {self.name} "
                                               f"({function.name}: {function.variable.shape}) does not match "
                                               f"the shape of the {repr(DEFAULT_VARIABLE)} specified for the "
                                               f"{repr(Mechanism.__name__)}.")

            # If length of Mechanism's variable is 1 but the function's is longer,
            #     reshape Mechanism's inner dimension to match function
            elif variable_len==1 and function_variable_len>1:
                variable_shape = list(variable.shape)
                variable_shape[-1] = function_variable.shape[-1]
                # self.parameters.variable.default_value = np.zeros(tuple(variable_shape))
                variable = np.zeros(tuple(variable_shape))

            # IMPLEMENTATON NOTE:
            #    Don't worry about case in which length of function's variable is 1 and Mechanism's is > 1
            #    as the reshaping of the function's variable will be taken care of in _instantiate_function

        return super()._handle_default_variable(default_variable=variable,
                                                size=size,
                                                input_ports=input_ports,
                                                function=function,
                                                params=params)



from . import inputport
from . import modulatorysignals
from . import outputport
from . import parameterport
from . import port

from .inputport import *
from .modulatorysignals import *
from .outputport import *
from .parameterport import *
from .port import *

__all__ = list(inputport.__all__)
__all__.extend(modulatorysignals.__all__)
__all__.extend(outputport.__all__)
__all__.extend(parameterport.__all__)
__all__.extend(port.__all__)
